<?php
/**
 * SPDX-FileCopyrightText: 2024 Sascha Nitsch (grumpydeveloper) https://contentnation.net/@grumpydevelop
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * @author Yannis Vogel (vogeldevelopment)
 **/

namespace Federator\DIO;

/**
 * IO functions related to fedUsers
 */
class FedUser
{
    /**
     * add local user based on given user object received from remote service
     * @param \mysqli $dbh database handle
     * @param \Federator\Data\FedUser $user user object to use
     * @param string $_user user/profile name
     * @return void
     */
    protected static function addUserToDB($dbh, $user, $_user)
    {
        // check if it is timed out user
        $sql = 'select unix_timestamp(`validuntil`) from fedusers where id=?';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError('FedUser::addLocalUser Failed to prepare statement');
        }
        $stmt->bind_param('s', $_user);
        $validuntil = 0;
        $ret = $stmt->bind_result($validuntil);
        $stmt->execute();
        if ($ret) {
            $stmt->fetch();
        }
        $stmt->close();
        if ($validuntil == 0) {
            $sql = 'insert into fedusers (id, url, name, publickey, summary, type, inboxurl, sharedinboxurl,';
            $sql .= ' followersurl, followingurl, publickeyid, outboxurl, validuntil)';
            $sql .= ' values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, now() + interval 1 day)';
            $stmt = $dbh->prepare($sql);
            if ($stmt === false) {
                throw new \Federator\Exceptions\ServerError('FedUser::addLocalUser Failed to prepare create statement');
            }
            $stmt->bind_param(
                'ssssssssssss',
                $_user,
                $user->actorURL,
                $user->name,
                $user->publicKey,
                $user->summary,
                $user->type,
                $user->inboxURL,
                $user->sharedInboxURL,
                $user->followersURL,
                $user->followingURL,
                $user->publicKeyId,
                $user->outboxURL
            );
        } else {
            // update to existing user
            $sql = 'update fedusers set validuntil=now() + interval 1 day, url=?, name=?, publickey=?, summary=?,';
            $sql .= ' type=?, inboxurl=?, sharedinboxurl=?, followersurl=?, followingurl=?, publickeyid=?, outboxurl=?';
            $sql .= ' where id=?';
            $stmt = $dbh->prepare($sql);
            if ($stmt === false) {
                throw new \Federator\Exceptions\ServerError('FedUser::extendUser Failed to prepare update statement');
            }
            $stmt->bind_param(
                'ssssssssssss',
                $user->actorURL,
                $user->name,
                $user->publicKey,
                $user->summary,
                $user->type,
                $user->inboxURL,
                $user->sharedInboxURL,
                $user->followersURL,
                $user->followingURL,
                $user->publicKeyId,
                $user->outboxURL,
                $_user
            );
        }
        try {
            $stmt->execute();
            $stmt->close();
            $user->id = $_user;
        } catch (\mysqli_sql_exception $e) {
            error_log($sql);
            error_log(print_r($user, true));
            error_log($e->getMessage());
        }
    }

    /**
     * extend the given user with internal data
     * @param \mysqli $dbh database  handle
     * @param \Federator\Data\FedUser $user user to extend
     * @param string $_user user/profile name
     */
    protected static function extendUser(\mysqli $dbh, \Federator\Data\FedUser $user, $_user): void
    {
        $sql = 'select id,unix_timestamp(`validuntil`) from fedusers where id=?';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError('FedUser::extendUser Failed to prepare statement');
        }
        $stmt->bind_param('s', $_user);
        $validuntil = 0;
        $ret = $stmt->bind_result($user->id, $validuntil);
        $stmt->execute();
        if ($ret) {
            $stmt->fetch();
        }
        $stmt->close();
        // if a new user, create own database entry with additionally needed info
        if ($user->id === null || $validuntil < time()) {
            self::addUserToDB($dbh, $user, $_user);
        }

        // no further processing for now
    }

    /**
     * get user by name
     *
     * @param \mysqli $dbh
     *          database handle
     * @param string $_name
     *          user name
     * @param \Federator\Cache\Cache|null $cache
     *          optional caching service
     * @return \Federator\Data\FedUser
     */
    public static function getUserByName($dbh, $_name, $cache)
    {
        $user = false;

        // ask cache
        if ($cache !== null) {
            $user = $cache->getRemoteFedUserByName($_name);
        }
        if ($user !== false) {
            return $user;
        }
        // check our db
        $sql = 'select `id`, `url`, `name`, `publickey`, `summary`, `type`, `inboxurl`, `sharedinboxurl`, ';
        $sql .= '`followersurl`, `followingurl`, `publickeyid`, `outboxurl`';
        $sql .= ' from fedusers where `id`=? and `validuntil`>=now()';
        $stmt = $dbh->prepare($sql);
        if ($stmt === false) {
            throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to prepare statement');
        }
        $stmt->bind_param('s', $_name);
        $user = new \Federator\Data\FedUser();
        $ret = $stmt->bind_result(
            $user->id,
            $user->actorURL,
            $user->name,
            $user->publicKey,
            $user->summary,
            $user->type,
            $user->inboxURL,
            $user->sharedInboxURL,
            $user->followersURL,
            $user->followingURL,
            $user->publicKeyId,
            $user->outboxURL
        );
        $stmt->execute();
        if ($ret) {
            $stmt->fetch();
        }
        $stmt->close();

        if ($user->id === null) {
            // check if its a federated user with username@domain.ending
            if (preg_match("/^([^@]+)@(.*)$/", $_name, $matches) == 1) {
                // make webfinger request
                $remoteURL = 'https://' . $matches[2] . '/.well-known/webfinger?resource=acct:' . urlencode($_name);
                $headers = ['Accept: application/activity+json'];
                [$response, $info] = \Federator\Main::getFromRemote($remoteURL, $headers);
                if ($info['http_code'] != 200) {
                    throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to fetch webfinger for '
                        . $_name);
                }
                $r = json_decode($response, true);
                if ($r === false || $r === null || !is_array($r)) {
                    throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to decode webfinger for '
                        . $_name);
                }
                // get the webwinger user url and fetch the user
                if (isset($r['links'])) {
                    foreach ($r['links'] as $link) {
                        if (isset($link['rel']) && $link['rel'] === 'self') {
                            $remoteURL = $link['href'];
                            break;
                        }
                    }
                }
                if (!isset($remoteURL)) {
                    throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to find self link '
                        . 'in webfinger for ' . $_name);
                }
            } else {
                $remoteURL = $_name;
            }
            // fetch the user
            $headers = ['Accept: application/activity+json'];
            [$response, $info] = \Federator\Main::getFromRemote($remoteURL, $headers);
            if ($info['http_code'] != 200) {
                throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to fetch user from '
                    . 'remoteUrl for ' . $_name);
            }
            $r = json_decode($response, true);
            if ($r === false || $r === null || !is_array($r)) {
                throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to decode user for '
                    . $_name);
            }
            $r['publicKeyId'] = $r['publicKey']['id'];
            $r['publicKey'] = $r['publicKey']['publicKeyPem'];
            if (isset($r['endpoints'])) {
                if (isset($r['endpoints']['sharedInbox'])) {
                    $r['sharedInbox'] = $r['endpoints']['sharedInbox'];
                }
            }
            $r['actorURL'] = $remoteURL;
            $data = json_encode($r);
            if ($data === false) {
                throw new \Federator\Exceptions\ServerError('FedUser::getUserByName Failed to encode userdata '
                    . $_name);
            }
            $user = \Federator\Data\FedUser::createFromJson($data);
        }

        if ($cache !== null && $user !== false) {
            if ($user->id !== null && $user->actorURL !== null) {
                self::addUserToDB($dbh, $user, $_name);
            }
            $cache->saveRemoteFedUserByName($_name, $user);
        }
        if ($user === false) {
            throw new \Federator\Exceptions\ServerError('FedUser::getUserByName User not found');
        }
        return $user;
    }

    /**
     * handle post call for specific user
     *
     * @param \Federator\Main $main main instance
     * @param \mysqli $dbh database handle
     * @param \Federator\Connector\Connector $connector connector to use
     * @param \Federator\Cache\Cache|null $cache optional caching service
     * @param string $_user user that triggered the post
     * @param string $_recipientId recipient of the post
     * @param \Federator\Data\ActivityPub\Common\Activity $inboxActivity the activity that we received
     * @return boolean response
     */
    public static function inboxForUser($main, $dbh, $connector, $cache, $_user, $_recipientId, $inboxActivity)
    {
        if (!isset($_user)) {
            error_log('Inbox::postForUser no user given');
            return false;
        }

        // get sender
        $user = \Federator\DIO\FedUser::getUserByName(
            $dbh,
            $_user,
            $cache
        );
        if ($user === null || $user->id === null) {
            error_log('Inbox::postForUser couldn\'t find user: ' . $_user);
            return false;
        }

        $type = strtolower($inboxActivity->getType());

        if ($_recipientId === '') {
            if ($type === 'undo' || $type === 'delete') {
                switch ($type) {
                    case 'delete':
                        // Delete Note/Post
                        $object = $inboxActivity->getObject();
                        if (is_string($object)) {
                            \Federator\DIO\Posts::deletePost($dbh, $object);
                        } elseif (is_object($object)) {
                            $objectId = $object->getID();
                            \Federator\DIO\Posts::deletePost($dbh, $objectId);
                        } else {
                            error_log('Inbox::postForUser Error in Delete Post for user ' . $user->id
                                . ', object is not a string or object');
                            error_log(' object of type ' . gettype($object));
                            return false;
                        }
                        break;

                    case 'undo':
                        $object = $inboxActivity->getObject();
                        if (is_object($object)) {
                            switch (strtolower($object->getType())) {
                                case 'like':
                                case 'dislike':
                                    // Undo Like/Dislike (remove like/dislike)
                                    $targetId = $object->getID();
                                    // \Federator\DIO\Votes::removeVote($dbh, $user->id, $targetId, 'dislike');
                                    \Federator\DIO\Posts::deletePost($dbh, $targetId);
                                    break;
                                case 'note':
                                case 'article':
                                    // Undo Note (remove note)
                                    $noteId = $object->getID();
                                    \Federator\DIO\Posts::deletePost($dbh, $noteId);
                                    break;
                            }
                        }
                        break;

                    default:
                        error_log('Inbox::postForUser Unhandled activity type ' . $type . ' for user ' . $user->id);
                        break;
                }

                return true;
            }
        }

        $atPos = strpos($_recipientId, '@');
        if ($atPos !== false) {
            $_recipientId = substr($_recipientId, 0, $atPos);
        }

        // get recipient
        $recipient = \Federator\DIO\User::getUserByName(
            $dbh,
            $_recipientId,
            $connector,
            $cache
        );
        if ($recipient === null || $recipient->id === null) {
            error_log('Inbox::postForUser couldn\'t find recipient: ' . $_recipientId);
            return false;
        }

        $rootDir = $_SERVER['DOCUMENT_ROOT'] .  '../';
        // Save the raw input and parsed JSON to a file for inspection
        file_put_contents(
            $rootDir . 'logs/inbox_' . $recipient->id . '.log',
            date('Y-m-d H:i:s') . ": ==== POST " . $recipient->id . " Inbox Activity ====\n"
                . json_encode($inboxActivity, JSON_UNESCAPED_SLASHES | JSON_PRETTY_PRINT) . "\n\n",
            FILE_APPEND
        );

        switch ($type) {
            case 'follow':
                $success = \Federator\DIO\Followers::addExternalFollow(
                    $dbh,
                    $inboxActivity->getID(),
                    $user->id,
                    $recipient->id
                );

                if ($success === true) {
                    // send accept back
                    $accept = new \Federator\Data\ActivityPub\Common\Accept();
                    $local = $inboxActivity->getObject();
                    if (is_string($local)) {
                        $accept->setAActor($local);
                        $id = bin2hex(openssl_random_pseudo_bytes(4));
                        $accept->setID($local . '#accepts/follows/' . $id);
                        $obj = new \Federator\Data\ActivityPub\Common\Activity($inboxActivity->getType());
                        $config = $main->getConfig();
                        $ourhost = $config['generic']['protocol'] . '://' . $config['generic']['externaldomain'];
                        $obj->setID($ourhost . '/' . $id);
                        $obj->setAActor($inboxActivity->getAActor());
                        $obj->setObject($local);
                        $accept->setObject($obj);
                        // send
                        \Federator\DIO\Server::sendActivity($dbh, $ourhost, $recipient, $user, $accept);
                    }
                } else {
                    error_log('Inbox::postForUser Failed to add follower for user ' . $user->id);
                }
                break;

            case 'delete':
                // Delete Note/Post
                $object = $inboxActivity->getObject();
                if (is_string($object)) {
                    \Federator\DIO\Posts::deletePost($dbh, $object);
                } elseif (is_object($object)) {
                    $objectId = $object->getID();
                    \Federator\DIO\Posts::deletePost($dbh, $objectId);
                }
                break;

            case 'undo':
                $object = $inboxActivity->getObject();
                if (is_object($object)) {
                    switch (strtolower($object->getType())) {
                        case 'follow':
                            $success = false;
                            if ($object instanceof \Federator\Data\ActivityPub\Common\Activity) {
                                $actor = $object->getAActor();
                                if ($actor !== '') {
                                    $success = \Federator\DIO\Followers::removeFollow($dbh, $user->id, $recipient->id);
                                }
                            }
                            if ($success === false) {
                                error_log('Inbox::postForUser Failed to remove follower for user ' . $user->id);
                            }
                            break;
                        case 'like':
                        case 'dislike':
                            // Undo Like/Dislike (remove like/dislike)
                            $targetId = $object->getID();
                            \Federator\DIO\Votes::removeVote($dbh, $user->id, $targetId);
                            // \Federator\DIO\Posts::deletePost($dbh, $targetId);
                            break;
                        case 'note':
                            // Undo Note (remove note)
                            $noteId = $object->getID();
                            \Federator\DIO\Posts::deletePost($dbh, $noteId);
                            break;
                    }
                }
                break;

            case 'like':
            case 'dislike':
                // Add Like/Dislike
                $targetId = $inboxActivity->getObject();
                if (is_string($targetId)) {
                    \Federator\DIO\Votes::addVote($dbh, $user->id, $targetId, $type);
                } else {
                    error_log('Inbox::postForUser Error in Add Like/Dislike for user ' . $user->id
                        . ', targetId is not a string');
                    return false;
                }
                break;

            case 'create':
            case 'update':
                $object = $inboxActivity->getObject();
                if (is_object($object)) {
                    switch (strtolower($object->getType())) {
                        case 'note':
                            \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                            break;
                        case 'article':
                            \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                            break;
                        default:
                            \Federator\DIO\Posts::savePost($dbh, $user->id, $inboxActivity);
                            break;
                    }
                }
                break;
            default:
                error_log('Inbox::postForUser Unhandled activity type $type for user ' . $user->id);
                break;
        }
        return true;
    }
}
